# -*- coding: utf-8 -*-
# This calss represents a Chromosome object

from matplotlib.patches import Wedge, Rectangle
from buscoplotpy.graphics.label import Label


class Chromosome():

    def __init__(self, x_start: float, 
                       x_end: float, 
                       y_start: float, 
                       y_end: float, 
                       horizontal: bool = True,
                       round_edges: bool = False
    ):
		
        """
        Initialize the Chromosome class.
        
        Args:
            x_start (float): The starting x-coordinate of the chromosome.
            x_end (float): The ending x-coordinate of the chromosome.
            y_start (float): The starting y-coordinate of the chromosome.
            y_end (float): The ending y-coordinate of the chromosome.
            horizontal (bool, optional): Whether the chromosome is horizontal. Defaults to True.
            round_edges (bool, optional): Whether the chromosome has rounded edges. Defaults to False.
        """

        # Coordinates
        self.x_start = x_start
        self.x_end   = x_end
        self.y_start = y_start
        self.y_end   = y_end
        
        # Graphical attributes
        self.horizontal  = horizontal
        self.round_edges = round_edges

        # Elements
        self.labels  = []
        self.regions = []
        
        if self.horizontal and self.round_edges:

            # Calculate the center and radius of the chromosome semicircles
            center_y = (self.y_end + self.y_start)/2.0
            radius   = (self.y_end - self.y_start)/2.0

            # Calculate the angles of the semicircles
            theta1   = -90.0
            theta2   =  90.0
            
            # Create the start and the end semicircles
            self.w1 = Wedge((self.x_start, center_y), radius, theta2, theta1, width=0.00001, facecolor='white', edgecolor='black', linewidth=0.6)
            self.w2 = Wedge((self.x_end,   center_y), radius, theta1, theta2, width=0.00001, facecolor='white', edgecolor='black', linewidth=0.6)

        elif not self.horizontal and self.round_edges:

            # Calculate the center and radius of the chromosome semicircles
            center_x = (self.x_end + self.x_start)/2.0
            radius   = (self.x_end - self.x_start)/2.0

            # Calculate the angles of the semicircles
            theta1   = -90.0
            theta2   =  90.0
            
            # Create the start and the end semicircles
            self.w1 = Wedge((center_x, self.y_start), radius, theta2, theta1, width=0.00001, facecolor='white', edgecolor='black', linewidth=0.6)
            self.w2 = Wedge((center_x, self.y_end),   radius, theta1, theta2, width=0.00001, facecolor='white', edgecolor='black', linewidth=0.6)
	
    def __str__(self):

        """
        Returns a string representation of the Chromosome object.
        
        Returns:
            str: The string representation of the object.
        """

        return 'Chromosome({}, {}, {}, {})'.format(self.x_start, self.x_end, self.y_start, self.y_end)

    def plot(self, ax, round_edges=False):

        """
        Plot the chromosome on the given axes.

        Parameters:
            - ax: The axes to plot on.
            - round_edges: A boolean indicating whether to round the edges of the chromosome.
        Returns:
            None
        """

        # Add the chromosome horizontal lines 
        ax.plot([self.x_start, self.x_end], [self.y_start, self.y_start], ls='-', color='black', linewidth=1)
        ax.plot([self.x_start, self.x_end], [self.y_end,   self.y_end],   ls='-', color='black', linewidth=1)

        if round_edges:

            # Add the semicircles on chromosomes
            ax.add_patch(self.w1)
            ax.add_patch(self.w2)	
        
        else:
        
            # Add the chromosome vertical lines
            ax.plot([self.x_start, self.x_start], [self.y_start, self.y_end], ls='-', color='black', linewidth=1)
            ax.plot([self.x_end,   self.x_end],   [self.y_start, self.y_end], ls='-', color='black', linewidth=1)
        
        # Plot the labels
        for label in self.labels:
            label.plot(ax)
        
        # Plot the regions
        for region in self.regions:
            ax.add_patch(region)

    def add_label(self, x: float, y: float, text: str, ha: str = 'left', va: str = 'baseline'):

        """
        Add a label to the plot at the specified coordinates.
        Args:
            x (float): The x-coordinate of the label.
            y (float): The y-coordinate of the label.
            text (str): The text of the label.
            ha (str, optional): The horizontal alignment of the label. Defaults to 'left'.
            va (str, optional): The vertical alignment of the label. Defaults to 'baseline'.
        """

        label = Label(x, y, text, ha=ha, va=va)
        self.labels.append(label)

    def add_region(self, anchor_point: (float, float), width: float, height: float, color: str = 'black', linewidth: int = 1):

        """
        Adds a region to the list of regions.
        
        Parameters:
            anchor_point (Tuple[float, float]): The anchor point of the region.
            width (float): The width of the region.
            height (float): The height of the region.
            color (str, optional): The color of the region. Defaults to 'black'.
            linewidth (int, optional): The linewidth of the region. Defaults to 1.
        """

        region = Rectangle(xy=anchor_point, width=width, height=height, color=color, linewidth=linewidth)
        self.regions.append(region)



import json
import pandas as pd

def load_json_summary(path: str) -> pd.DataFrame:

    """
    Load the summary file generated by BUSCO into a pandas DataFrame.
    
    Parameters:
        path (str): The path to the summary file.
        
    Returns:
        pd.DataFrame: The loaded summary table with busco gene information.
    """
    
    # Read the summary table from the file
    dict = json.load(open(path))

    # Populate the columns of the summary table with data from the dict
    summary_table = pd.DataFrame({
        'max_intron': dict['parameters']['max_intron'],
        'max_seq_len': dict['parameters']['max_seq_len'],
        'metaeuk_parameters': dict['parameters']['metaeuk_parameters'],
        "metaeuk_rerun_parameters": dict['parameters']['metaeuk_rerun_parameters'],
        "contig_break": dict['parameters']['contig_break'],
        "scaffold_composition": dict['parameters']['scaffold_composition'],
        "gene_predictor": dict['parameters']['gene_predictor'],
        'dataset_name': dict['lineage_dataset']['name'],
        'number_of_buscos': dict['lineage_dataset']['number_of_buscos'],
        'number_of_species': dict['lineage_dataset']['number_of_species'],
        'metaeuk_version': dict['versions']['metaeuk'],
        'bbtools_version': dict['versions']['bbtools'],
        'busco_version': dict['versions']['busco'],
        'hmmsearch_version': dict['versions']['hmmsearch'],
        'one_line_summary': dict['results']['one_line_summary'],
        'complete': dict['results']['Complete'],
        'single copy': dict['results']['Single copy'],
        'multi copy': dict['results']['Multi copy'],
        'fragmented': dict['results']['Fragmented'],
        'missing': dict['results']['Missing'],
        'n_markers': dict['results']['n_markers'],
        'domain': dict['results']['domain'],
        'number_of_scaffolds': dict['results']['Number of scaffolds'],
        'number_of_contigs': dict['results']['Number of contigs'],
        'total length': dict['results']['Total length'],
        'percent gaps': dict['results']['Percent gaps'],
        'scaffold N50': dict['results']['Scaffold N50'],
        'contigs N50': dict['results']['Contigs N50']
    }, index=[0])

    return summary_table


#Importing libraries
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import pandas as pd

from matplotlib.patches import Patch

def organism_busco_barplot(df: pd.DataFrame,
                           group_name: str = '',
                           organism_name: str = '',
                           out_path: str = './', 
                           filename: str = 'busco_barplot',
                           dpi: int = 300,
                           plt_show: bool = False
                        ) -> None:
    
    """
    Generate a barplot to visualize the completeness of assembly for different organisms on the same BUSCO dataset.

    Parameters:
        - df (pd.DataFrame): The input dataframe containing the data to be plotted.
        - group_name (str): The name of the group to be plotted.
        - organism_name (str): The name of the organism to be plotted.
        - out_path (str): The path where the plot image will be saved.
        - filename (str): The name of the plot image file.

    Output:
        - A barplot image with the completeness of assembly for different organisms on the same BUSCO dataset in .png format.

    Returns:
        - None
    """

    if len(df) == 0:
        return None

    if group_name == '':
        group_name = df['group'].iloc[0]


    # Create a list of species names
    species_names = []

    # If the organism name is not provided, get the unique species names from the dataframe    
    if organism_name == '':
        organisms = df['organism'].unique()

        for organism in list(organisms):
            species_names = species_names + [organism + '_' + i for i in df[df['organism'] == organism]['version']]
    
    # Else if the organism name is provided, set the list of species names to the provided name
    else:
        species_names = [organism_name]
    
    # Get the group and dataset name from the dataframe
    dataset_name = df['dataset_name'].iloc[0]

    # Create a matrix of completeness values
    matrix = df[['single copy', 'multi copy', 'fragmented', 'missing']].to_numpy()

    # Get the one-line summary
    one_line_summary = df['one_line_summary'].to_list()

    # Set the style of Seaborn
    sns.set(style="whitegrid")

    # Define the colors and labels for the plot
    colors = ['#49a34b', '#636633', '#664f33', '#4b5669']
    labels = ['Complete - single', 'Complete - multi', 'Fragmented', 'Missing']
    busco_labels = ['Single copy', 'Multi copy', 'Fragmented', 'Missing']

    # Generate the values matrix
    values_matrix = matrix.copy()

    # Calculate the cumulative sum of the matrix
    for row in range(len(values_matrix)):
        for col in range(len(values_matrix[row])):
            if col != 0:
                values_matrix[row][col] = values_matrix[row][col-1] + values_matrix[row][col]

    # Transpose the matrix
    values_matrix = values_matrix.T

    # Create the plot
    fig, axs = plt.subplots(figsize=(20, len(species_names)+1), ncols=1, nrows=1)
    
    # Create the individual bars with different colors
    for idx, label in enumerate(labels):
        sns.barplot(x=values_matrix[-idx-1], color=colors[-idx-1], y=species_names, ax=axs, errorbar=None, alpha=0.99)
    
    # Add a text string to each bar
    for i in range(len(one_line_summary)):
        axs.text(0.9, i+0.1, str(one_line_summary[i]), color='white', fontsize=15)
    
    # Set the locator of the major ticker
    axs.xaxis.set_major_locator(ticker.MultipleLocator(5))
    
    # Customize the x and y axes and tick labels
    axs.set_xlabel('Percentage', fontsize=18)
    axs.set_ylabel('Organism', fontsize=18)
    axs.set_title(dataset_name + ' ' + group_name + ' - Barplot of completeness of assembly', fontsize=20, weight='bold', pad=30)
    axs.tick_params(labelsize=15)

    # Set the x-axis limits
    axs.set_xlim(0, 100)

    # Remove the borders of the plot
    sns.despine(top=True, right=True, left=True, bottom=True)

    # Remove the y-axis ticks
    axs.tick_params(axis='y', length=0)

    # Create the legend
    patches_list = [Patch(color=colors[i], label=labels[i], alpha=0.75) for i in range(len(labels))]
    axs.legend(handles=patches_list, bbox_to_anchor=(-0.5, 1), fontsize=15)

    
    # Create the values table
    #table_data = [["Species"] + busco_labels]
    #for row, species in zip(matrix, species_names):
    #    table_data.append([species] + [str(i) + '%' for i in row])
    #table = axs[1].table(cellText=table_data, loc='center')
    #table.set_fontsize(34)
    #table.scale(1, 2)
    #table.auto_set_column_width(col=range(len(table_data)))
    #axs[1].set_title("Percentage table", weight='bold', size=15)
    #for (row, col), cell in table.get_celld().items():
    #    if (row == 0) or (col == -1):
    #        cell.set_text_props(fontproperties=FontProperties(weight='bold'))  
    #axs[1].axis('off')

    # Save and show the plot
    plt.savefig(out_path + filename + '_completeness.png', bbox_inches='tight', dpi=dpi)

    if plt_show:
        plt.show()
    
    plt.close()


# -*- coding: utf-8 -*-

#Importing libraries
import matplotlib.pyplot as plt
import pandas as pd
from buscoplotpy.graphics.chromosome import Chromosome

from matplotlib.patches import Wedge, Rectangle

def karyoplot(karyotype_file: str, 
              output_file: str = '', 
              title: str = 'Karyoplot', 
              fulltable: pd.DataFrame = None, 
              dpi: int = 300, 
              chrs_limit: int = 30, 
              plt_show: bool = False,
              palette: str in ['green', 'azure'] = 'green',
              bbox_inches: str = 'tight',
              dim: int = 2
) -> None:

    """
    Plot a karyotype based on the karyotype file and the BUSCO fulltable.

    Parameters:
        karyotype_file (str): The path to the karyotype file.
        output_file (str, optional): The path to save the output plot.
        title (str, optional): The title of the plot.
        fulltable (pd.DataFrame): The BUSCO's full table DataFrame.
        dpi (int, optional): The DPI (dots per inch) of the output plot. Default is 300.
        chrs_limit (int, optional): The maximum number of chromosomes to plot. Default is 30.
        plt_show (bool, optional): Whether to show the plot. Default is False.

    Output:
        - The karyotype plot in png format.
    Returns:
        None
    """

    # Define the colors
    green = ['green', 'gray', 'black']
    azure = ['#5999ff', '#ffff05', '#21211d']
    selected = []
    
    # Selecting the right palette
    if palette == 'green':
        selected = green
    elif palette == 'azure':
        selected = azure

    def get_color(status: str, palette: str = 'green') -> str:

        """
        Get the color based on the status of the item.
        Parameters:
            item (dict): The item containing the status.
        Returns:
            str: The color corresponding to the status.
        """
        
        # Reurn the color based on the palette/status
        if status == 'Complete':
            return selected[0]
        elif status == 'Duplicated':
            return selected[1]
        else:
            return selected[2]

    # Read the karyotype file into a DataFrame
    karyotype = pd.read_csv(karyotype_file, sep="\t")

    # Lowercase the column names
    karyotype.columns = karyotype.columns.str.lower()

    # Remove rows where status is 'Missing'
    fulltable = fulltable[fulltable['status'] != 'Missing']

    # Extract the sequence name from the 'sequence' column
    fulltable.loc[:, 'sequence'] = fulltable['sequence'].map(lambda x: x.split(':')[0])

    # If the number of chromosomes is greater than chr_limit,
    #   then select the most significant chromosomes

    if len(karyotype) > chrs_limit:

        if len(fulltable) == 0:
            karyotype = karyotype.iloc[:chrs_limit, :]
        else:
            karyotype.set_index('chr', inplace=True)

            # Select the most significant chromosomes (the chromosomes with more hits)
            first_chrs = fulltable['sequence'].value_counts().index.to_list()[:chrs_limit]
            karyotype = karyotype.loc[first_chrs].sort_values(by='end', ascending=False)
            karyotype = karyotype.reset_index()

    # Calculate the limits of the plot
    X_lim = 100
    Y_lim = dim * len(karyotype) + dim + 5

    # Create a new figure and axis
    fig, ax   = plt.subplots(figsize=(20, 20*Y_lim/X_lim), dpi=dpi)

    # Turn off the axis
    ax.axis('off')

    # Set the x and y limits of the plot
    ax.set_xlim([0, X_lim])
    ax.set_ylim([0, Y_lim])

    # Insert the plot title
    ax.text(X_lim / 2, Y_lim - 1, title, fontsize=20, ha='center')

    # Calculate the maximum length of the chromosome name
    chr_max_len = len(max(karyotype['chr'], key=len))

    # Get the maximum length of the chromosome
    chr_max_dim = karyotype['end'].max()

    # Plot the karyotypes
    for index, row in karyotype.iterrows():
        
        # Get the dimension of the karyotype
        chr_dim = row['end']

        # Define the coordinates for the rectangle
        x_start = chr_max_len / 2
        x_end   = x_start + chr_dim * (X_lim*9/10) / chr_max_dim
        y_start = (len(karyotype) - index) * dim
        y_end   = y_start + dim / 2

        C = Chromosome(x_start, x_end, y_start, y_end)
        C.add_label(x=0.0, y=(y_start + y_end) / 2, text=row['chr'], horizontalalignment='center', verticalalignment='center')
        

        # Create all chromosome region
        for i, item in fulltable[fulltable['sequence'] == row.chr].iterrows():

            # Define the coordinates for the region
            converted_start_pos = item['gene_start'] * (x_end - x_start) / chr_dim
            converted_end_pos   = item['gene_end']   * (x_end - x_start) / chr_dim

            anchor_point = (x_start + converted_start_pos, y_start)
            width        = converted_end_pos - converted_start_pos
            height       = y_end - y_start

            #:                +------------------+
            #:                |                  |
            #:              height               |
            #:                |                  |
            #:               (xy)---- width -----+

            # Add the seleted region to the chromosome
            C.add_region(anchor_point=anchor_point, width=width, height=height, color=get_color(item['status'], palette), linewidth=1)
        
        # Plot the chromosome
        C.plot(ax)

    # Write the legend
    plt.legend(handles=[Rectangle((0,0),1,1, color=selected[0]), 
                        Rectangle((0,0),1,1, color=selected[1]), 
                        Rectangle((0,0),1,1, color=selected[2])],
                        labels=['Complete', 'Duplicated', 'Fragmented'], 
                        loc='upper right'
    )

    # Save and show the plot
    if output_file:
        plt.savefig(output_file, dpi=dpi, bbox_inches=bbox_inches)

    if plt_show:
        plt.show()
    
    plt.close()



#Importing libraries
import pandas as pd

def load_busco_fulltable(path: str, 
                         group: str='', 
                         organism: str='', 
                         genome_version: str=''
) -> pd.DataFrame:

    """
    Load the full table generated by BUSCO into a pandas DataFrame.
    
    Parameters:
        group (str): The group the organism belongs to.
        organism (str): The name of the organism.
        genome_version (str): The version of the genome.
        path (str): The path to the full table.
        
    Returns:
        pd.DataFrame: The loaded full table with busco gene information.
    """
    
    # Initialize the empty full table
    full_table = pd.DataFrame()

    # Read the busco table from the file
    busco_table = pd.read_csv(path, skiprows=2, sep='\t')

    # Populate the columns of the full table with data from the busco table
    full_table['busco_id'] = busco_table['# Busco id']
    full_table['status'] = busco_table['Status']
    full_table['sequence'] = busco_table['Sequence']
    full_table['gene_start'] = busco_table['Gene Start']
    full_table['gene_end'] = busco_table['Gene End']
    full_table['strand'] = busco_table['Strand']
    full_table['score'] = busco_table['Score']
    full_table['length'] = busco_table['Length']
    full_table['group'] = group
    full_table['organism'] = organism
    full_table['genome_version'] = genome_version

    # Try to assign the 'OrthoDB url' column to the full table, otherwise assign None
    try:
        full_table['ortho_db_url'] = busco_table['OrthoDB url']
    except:
        full_table['ortho_db_url'] = None
    
    # Try to assign the 'Description' column to the full table, otherwise assign None
    try:
        full_table['description'] = busco_table['Description']
    except:
        full_table['description'] = None

    # Return the loaded full table
    return full_table


# -*- coding: utf-8 -*-
# This class represents a label on a plot.

class Label():

    def __init__(self, x: float, y: float, text: str, ha: str = 'left', va: str = 'baseline'):

        """
        Initialize the object with the given values.
        Args:
            x (float): The x-coordinate of the object.
            y (float): The y-coordinate of the object.
            text (str): The text content of the object.
            ha (str): The horizontal alignment of the object. Default is 'left'.
            va (str): The vertical alignment of the object. Default is 'baseline'.
        """
        
        # Coordinates
        self.x    = x
        self.y    = y
        self.text = text

        # Graphical attributes
        self.horizontalalignment = ha
        self.verticalalignment   = va
    
    def plot(self, ax):

        """
        Plot the text on the given axes.
        Parameters:
            ax (matplotlib.axes.Axes): The axes on which to plot the text.
        """

        # Add the text to the axes at the specified coordinates
        ax.text(self.x, self.y, self.text, ha=self.horizontalalignment, va=self.verticalalignment)